# Contribution Guide for Multiphase Cases Repository by HZDR

[[_TOC_]]

## General Remarks

In general we adopt the code style guidelines defined by the OpenFOAM Foundation

* <http://openfoam.org/dev/how-to-contribute>
* <http://openfoam.org/dev/coding-style-guide>
* <https://openfoam.org/dev/code-quality>

Make sure that no source code or utilities of the OpenFOAM Foundation software
are copied to your case, because it is licensed under GNU General Public License
and the content in the repository has a different [license](LICENSE).

In case you need to store larger binary or ascii data, e.g. tabbed data to
set-up your case, consider using `gzip` to compress them. OpenFOAM is capable of
using `*.gz` files as input, e.g. as initial condition.

## Static Code Analysis (Linting)

Make sure that dictionary files contain a valid header if they are originally
developed for the addon. You can find examples:

* for dictionaries: `etc/controlDict`
* for Bash: `etc/script/bash.sh`
* for Python: `etc/script/python.py`
* for Snakemake: `etc/script/Snakemake`

### Linting Code Style

Use the `interactive-hook` in `bin/tools` to check for tabs, trailing
whitespaces, and date in copyright headers

```shell
$WM_PROJECT_USER_DIR/bin/tools/interactive-hook -s HEAD -c file1 file2 ...
```

Furthermore, a `pre-commit-hook` is available in
`$WM_PROJECT_USER_DIR/bin/tools`

```shell
ln -s $WM_PROJECT_USER_DIR/bin/tools/pre-commit-hook .git/hooks/pre-commit
```

### Linting Python, Bash, Markdown, and other

MegaLinter[^1] is used for linting various formats like Python, Shell, JSON,
Markdown and other. To run MegaLinter locally use

```shell
docker run --rm --volume "$PWD:/tmp/lint" registry.hzdr.de/fwdc/infrastructure/megalinter:latest
```

Spellchecking is done with `cspell`[^2] for all strings and comments (included
in MegaLinter). If variables are used inside strings or comments or in the
documentation header they have to be added to the `cspell` dictionary by editing
`etc/cspell-dictionary.txt`. The dictionary can be sorted alphabetically by

```shell
sort cspell-dictionary.txt -o cspell-dictionary.txt
```

## Recommended Case Structure

The following points are a recommendation, but depending on the case
variations are possible.

### Simulation Setup

#### Case structure

A case should:

* be named according to the literature reference
* follow the case structure `etc/templatedCase`
* contain a sub-directory called `validation` for the validation data
* within the `validation` directory contain a sub-directory `reference` for the
  reference solution,
* and a sub-directory `experiment` for the experimental data

In case generated output files which are required for validation contain tabs
these can be removed by 1) `sed -i 's/\t/ /g' file`, or
2) `column -t file | tee file` to keep tables visually appealing.

Desired 80 characters in files, such as in a README.md, can be enforced via the
`fold -s file | tee file` command.

#### Pre-processing

Before contributing a case to be computed in parallel it is generally advised
to check for the optimal number of subdomains to reduce computational times. As
a rule of thumb around 10.000 cells per processor core should be used.

#### Post-processing

For post-processing:

* use gnuplot and/or python
* use runApplication ONLY for utilities and solvers
* generate graphs by `( cd validation && ./createGraphs )`
* add a `case.yml.orig` file for the metadata:
  * specify the relative path to the experimental
      data, reference solution and simulation result for validation, and
  * give a list of keywords describing the case according to the
      [Keyword List](etc/keywords.yml).

Note that in order for the goodness evaluation to work the names of the
validation fields given in the file headers for experimental data, reference
and current solution need to match and correspond to the field name given in
the `case.yml`. See the file headers in `line.csv` and `alpha.dat` of the case
template in `etc/templatedCase/template/validation` matching the naming of the
validation field of the result file to be created in the postProcessing folder.

A template for createGraphs can be found in
`etc/templatedCase/template/validation`.

The line configuration should be set according to other cases in the repo:

* green solid line for the reference solution
* black dashed line for the computed result
* red dots for the experimental data

See other cases for reference.

### Snakemake Workflow Setup

The workflow implementation assumes that each case archived in the repository
features scripts for

1. **configuration**, `Allconfigure`, e.g. setting parameters with
`foamDictionary`
2. **running**, `Allrun`, which gathers all commands for pre-processing,
solving and post-processing
3. **cleaning**, `Allclean`, i.e. to return to the original state of the case
4. **validation**, `Allvalidate`, e.g. comparing results against reference
solutions
5. **updating reference solutions**, `Allupdate`
6. **executing goodness evaluation**, `Allfuzzy`

While this is the recommended structure for efficiently using the workflow to
maintain a validation data base, the only **mandatory scripts** are `Allrun`
and `Allclean`. If any of the other scripts are not present, they are simply
not executed during the corresponding step in the workflow.

---
[^1]: <https://oxsecurity.github.io/megalinter/latest>
[^2]: <https://cspell.org>
