#!/bin/bash
#---------------------------------*- sh -*-------------------------------------
# ==  == ====== ====   ====    |
#                   \\     ||  | Multiphase Cases Repository by HZDR
# ======   //   ||  || ===//   | Website: https://doi.org/10.14278/rodare.811
# ||  ||  //    ||  // || \\   | License: Creative Commons Attribution 4.0
# ==  == ====== ====   ==  ==  |
#------------------------------------------------------------------------------
# License
#     This file is part of the Multiphase Cases Repository by HZDR.
#
#     Copyright (C) 2024 by Helmholtz-Zentrum Dresden-Rossendorf e.V. (HZDR),
#     Website: https://hzdr.de
#
#     The Multiphase Cases Repository by HZDR is licensed under the
#     Creative Commons Attribution 4.0 International License. To view a copy of
#     this license, visit http://creativecommons.org/licenses/by/4.0/ or send a
#     letter to Creative Commons, PO Box 1866, Mountain View, CA 94042, USA.
#
# Description
#     Create Snakefiles for cases to be incorporated in the workflow.
#
#------------------------------------------------------------------------------

PROGNAME=$(basename "$0")
FORCE=no

usage() {
   while [ "$#" -ge 1 ]; do echo "${1}"; shift; done
   cat<<USAGE

usage: $PROGNAME [OPTION] path setup case

options:
  -force        remove existing Snakefiles
  -help         this usage

* Create Snakefiles for cases to be incorporated in the workflow.

USAGE
   exit 1
}

# parse options
while [ "$#" -gt 0 ]
do
   case "${1}" in
   -h | -help)
      usage
      ;;
   -f | -force)
      FORCE=yes
      shift
      ;;
   -*)
      usage "Error: Invalid option '${1}'"
      ;;
   *)
      break
      ;;
   esac
done

path="$1"
setup="$2"
case="$3"
file=".snakefile"

# Remove old Snakefile if requested
if [ "$FORCE" == "yes" ]; then
    echo "Removing old $file"
    rm -f $file
fi

# Write Snakefile
rule_name=$(echo "${path}" | sed s/"\/"/"_"/g)
if [ -f $file ]; then
    echo "Skipping $file"
else
    echo "Creating $file"
    cat<<EOF>>"$file"
use rule Allconfigure_case as Allconfigure_${rule_name} with:
    output:
        '${path}/log.Allconfigure'

use rule Allrun_case as Allrun_${rule_name} with:
    input:
        '${path}/log.Allconfigure'
    output:
        '${path}/log.Allrun',
        report
        (
            directory('${path}/report'),
            patterns=['{name}.png'],
EOF
        if [ "${case}" == "" ]; then
            cat<<EOF>>"$file"
            category='${setup}'
EOF
        else
            cat<<EOF>>"$file"
            category='${setup}',
            subcategory='${case}'
EOF
        fi
    cat<<EOF>>"$file"
        )

rule premature_report_${rule_name}:
    output:
        report
        (
            directory('${path}/reportPremature'),
            patterns=['{name}.png'],
EOF
        if [ "${case}" == "" ]; then
            cat<<EOF>>"$file"
            category='${setup}'
EOF
        else
            cat<<EOF>>"$file"
            category='${setup}',
            subcategory='${case}'
EOF
        fi
    cat<<EOF>>"$file"
        )

use rule Allvalidate_case as Allvalidate_${rule_name} with:
    output:
        '${path}/log.Allvalidate'

use rule Allfuzzy_case as Allfuzzy_${rule_name} with:
    output:
        '${path}/log.Allfuzzy'

use rule Allupdate_case as Allupdate_${rule_name} with:
    output:
        '${path}/log.Allupdate'

use rule Allclean_case as Allclean_${rule_name} with:
    output:
        '${path}/log.Allclean'
EOF
fi

exit 0

#------------------------------------------------------------------------------
